/* **********************************************************************
 * File  : buf.c                                                        *
 * Usage : Dynamic buffer allocation module                             *
 * By    : Sean Butler (Xin)                                            *
 *                                                                      *
 *                              For use with Finality, Version 1.0      *
 *                                      by Rasdan and Xin               *
 * Original CircleMUD code by Jeremy Elsen, John Hopkins University     *
 ********************************************************************** */

#if 0

buf.c --  Implementation of a dynamically allocated text
buffering system.  Its intended use is to hold multiple
long strings, whose size is unknown. The following explains
a typical usage.

int my_function(int i)
{
   int i;
   BUF buf;
   buf = newbuf();
   for (i = 0; i < 1000; i++)
       buf = bufcat(buf, "Buffer test string\n");
   if (buf != NOT_ALLOCATED)
       i = strlen(buftos(buf));
   bfree(buf);
}
Note: bufcat() and bufcpy() will return -1 if unsuccessful
      the buffer index passed to them will be lost. Also note
      that it is neccessary to call bfree() for each buffer that
      you call newbuf() on.  Otherwise the data will hang around
      in memory and will be lost.

#endif

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include "buf.h"

int buflen(void);
int goodindex(BUF);

char terminator;
char *END = &terminator;   /* this value will mark list's end */

/* The list of buffers */
char **buf_list;

/* PUBLIC FUNCTIONS *************************************************/


/* This needs to be excuted at boot time.  (db.c)
 */
void initbufs()
{
    buf_list = (char **)malloc(sizeof(char *) * 1);
    *buf_list = END;
}

/* called to request a new buffer, returns an index of type BUF
 */
BUF newbuf(void)
{
    char **b;
    unsigned long length;
    for (b = buf_list; *b != END; b++)  /* find a free buffer */
        if (*b == NULL)
            break;
    if (*b == END) {                    /* end of the buffer list? */
        /* reallocate buffer list to include new one */
        length = b - buf_list + 2;  /* space for new buffer and the END marker */
        buf_list = (char **)realloc(buf_list, (size_t)(sizeof(char **) * length));
        b = buf_list + (size_t)(length - 1);
        *b = END;
        b--;
    }
    *b = (char *)malloc(sizeof(char) * ALLOC_SIZE); /* allocate the buffer */

    if (*b == NULL)
        return NOT_ALLOCATED;
    else {
        **b = '\0';          /* NULL terminate the buffer */
        return b - buf_list; /* return the index */
    }
}

/* buffer to string, returns a pointer the string that the
 * passed index refers to.
 */
char *buftos(BUF i)
{
     if (!goodindex(i))
        return NULL;
     else
        return *(buf_list + (size_t)i);
}

/* Copies a string into a buffer, reallocating the buffer if
 * neccessary.  Returns NOT_ALLOCATED if either the buffer
 * did not exist or could not be allocated.
 */
BUF bufcpy(BUF i, char *s)
{
    if (!goodindex(i))
        return NOT_ALLOCATED;
    **(buf_list + (size_t)i) = '\0';
    i = bufcat(i, s);
    return i;
}

/* Concatenates string s to the end of the buffer refered to by
 * by index.  Returns NOT_ALLOCATED if either the buffer
 * did not exist or could not be allocated.
 */
BUF bufcat(BUF i, char *s)
{
    int bufs_alloced;
    int buf_len;

    if (!goodindex(i))
        return NOT_ALLOCATED;
    buf_len = strlen(*(buf_list + (size_t)i));
    bufs_alloced = (buf_len / ALLOC_SIZE) + 1;

    if ((strlen(s) + buf_len) >= ((bufs_alloced * ALLOC_SIZE) - 1))
        *(buf_list + (size_t)i) = (char *)realloc( *(buf_list + (size_t)i),
                              (bufs_alloced + 1)* ALLOC_SIZE * sizeof(char));

    if (*(buf_list + (size_t)i) == NULL)
        return NOT_ALLOCATED;
    strcat(*(buf_list + (size_t)i), s);
    return i;

}

/* Works like sprintf, but there are some provisos...
 * 1.  If you need to add on more than 5000 chars in one call to bprintf,
 *     you will need to increase MAX_CAT below.
 * 2.  If you wish to use another buffer as a parameter you must
 *     convert it to char * using buftos() and use %s in the format string.
 *     NOTE: there is no standard way of handling string concatenation
 *           using sprintf and the like.  So you better be sure it works on
 *           your system before doing such operations.
 *
 */
#define MAX_CAT 5000

int bprintf(BUF b, const char *format, ...)
{
    va_list argptr;
    int cnt;
    char *tmp;

    tmp = (char *)malloc(sizeof(char) * strlen(buftos(b)) + MAX_CAT);

    va_start(argptr, format);
    cnt = vsprintf(tmp, format, argptr);
    va_end(argptr);

    if ( (cnt == EOF) || (bufcpy(b, tmp) == NOT_ALLOCATED) )
        cnt = EOF;

    free(tmp);
    return cnt;   
}


/* Frees the buffer refered to by i.
 * Does nothing if given an invalid buffer.
 */
void bfree(BUF i)
{
    if (!goodindex(i))
        return;

    free (*(buf_list + (size_t)i));
    *(buf_list + (size_t)i) = NULL;
}



/* INTERNAL FUNCTIONS **********************************/

/* counts the number of buffers that have been allocated.
 */
int buflen()
{
    int i;
    for (i = 0; *(buf_list + i) != END; i++)
        ;
    return i;
}

/* returns TRUE if the buffer exists and has been allocated
 * returns FALSE if the buffer does not exist or has not
 * been allocated.
 */
int goodindex(BUF i)
{
    if (i < 0)
       return FALSE;
    if (i >= buflen())
       return FALSE;
    if (*(buf_list + (size_t)i) == NULL)
       return FALSE;

    return TRUE;
}




